from warnings import warn
from contextlib import suppress
from typing import TYPE_CHECKING, Optional, Callable
from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import parse_qs

from liqpy.client import Client
from liqpy.dev import LiqpyWarning

if TYPE_CHECKING:
    from liqpy.types import LiqpayCallbackDict


class LiqpayHandler(BaseHTTPRequestHandler):
    """
    Liqpay HTTP request handler for testing

    Do not use in production!
    """

    server: "LiqpayServer"

    @property
    def client(self) -> "Client":
        return self.server.client

    def _is_url_encoded(self) -> bool:
        return self.headers.get("Content-Type", "").startswith(
            "application/x-www-form-urlencoded"
        )

    def _parse_body(self) -> tuple[str, str]:
        if not self._is_url_encoded():
            raise ValueError("Content-Type must be application/x-www-form-urlencoded")

        content_length = int(self.headers.get("Content-Length", -1))

        body = self.rfile.read(content_length).decode()
        result = parse_qs(body)

        signature = result["signature"][0]
        data = result["data"][0]

        return data, signature

    def _handle_webhook(self):
        data, signature = self._parse_body()
        return self.client.callback(data, signature, verify=self.server.verify)

    def do_POST(self):
        self.log_message('"POST %s %s"', self.path, self.protocol_version)
        try:
            self.server.callback(self._handle_webhook())
        finally:
            self.send_response(204)
            self.end_headers()


class LiqpayServer(HTTPServer):
    """
    Liqpay server for testing

    Do not use in production!
    """

    client: "Client"
    verify: bool
    callback: Callable[["LiqpayCallbackDict"], None]

    def __init__(
        self,
        /,
        host: str = "localhost",
        port: int = 8000,
        *,
        callback: Callable[["LiqpayCallbackDict"], None] = lambda _: None,
        client: Optional["Client"] = None,
        timeout: float | None = None,
        verify: bool = True,
    ):
        super().__init__((host, port), LiqpayHandler)
        self.client = Client() if client is None else client
        self.callback = callback
        self.verify = verify

        if timeout is not None:
            self.timeout = float(timeout)

        self.allow_reuse_address = True
        self.allow_reuse_port = True

        warn(
            "LiqPy Test Server is only for development and testing purposes. "
            "Do not use it in production!",
            category=LiqpyWarning,
            stacklevel=2,
        )

    def handle_callback(
        self, timeout: float | None = None
    ) -> Optional["LiqpayCallbackDict"]:
        result: Optional["LiqpayCallbackDict"] = None
        previous_callback = self.callback
        previous_timeout = self.timeout

        if timeout is not None:
            self.timeout = timeout

        def cb(value: "LiqpayCallbackDict"):
            nonlocal result
            result = value

        self.callback = cb
        self.handle_request()

        self.callback = previous_callback
        self.timeout = previous_timeout

        return result


if __name__ == "__main__":
    from logging import getLogger, basicConfig
    from pprint import pprint

    from dotenv import load_dotenv

    basicConfig(level="DEBUG")

    load_dotenv(".env")

    logger = getLogger()

    with LiqpayServer(client=Client(), callback=lambda c: pprint(c)) as server:
        host, port = server.server_address
        logger.info(f"LiqPy Test Server listening on {host}:{port}")

        with suppress(KeyboardInterrupt):
            server.serve_forever()

    logger.info(f"LiqPy Test Server closed")



from warnings import warn
from typing import Optional, Literal, Union, TYPE_CHECKING, Unpack, AnyStr, Type
from os import environ
from logging import getLogger
from datetime import datetime, timedelta
from numbers import Number
from re import search
from uuid import UUID

from httpx import Client as _Client, AsyncClient as _AsyncClient, Response
from secret_type import secret, Secret

from liqpy import __version__
from liqpy.dev import LiqpyWarning

from .api import (
    VERSION,
    Endpoint,
    post,
    sign,
    request,
    post_async,
    encode,
    decode,
    is_sandbox,
    exception,
    BasePreprocessor,
    BaseValidator,
    JSONEncoder,
    JSONDecoder,
    Decoder,
    Encoder,
    Preprocessor,
    Validator,
)

if TYPE_CHECKING:
    from .types.common import Language, Currency, SubscribePeriodicity, PayOption
    from .types.request import Format, LiqpayRequestDict, Action
    from .types.callback import LiqpayCallbackDict, LiqpayRefundDict


__all__ = ["BaseClient", "Client", "AsyncClient"]


logger = getLogger(__package__)


CHECKOUT_ACTIONS = (
    "auth",
    "pay",
    "hold",
    "subscribe",
    "paydonate",
)


class BaseClient:
    """
    Base class for LiqPay API logic that does not depend on the HTTP client.
    """
    _public_key: str
    _private_key: Secret[bytes]

    validator: BaseValidator
    preprocessor: BasePreprocessor
    encoder: JSONEncoder
    decoder: JSONDecoder

    def __init__(
        self,
        /,
        public_key: str | None = None,
        private_key: str | None = None,
        *,
        validator: Optional[BaseValidator] = None,
        preprocessor: Optional[BasePreprocessor] = None,
        encoder: Optional[JSONEncoder] = None,
        decoder: Optional[JSONDecoder] = None,
    ):
        self.update_keys(public_key=public_key, private_key=private_key)
        self.validator = validator if validator is not None else Validator()
        self.preprocessor = preprocessor if preprocessor is not None else Preprocessor()
        self.encoder = encoder if encoder is not None else Encoder()
        self.decoder = decoder if decoder is not None else Decoder()

    @property
    def public_key(self) -> str:
        """Public key used for requests"""
        return self._public_key

    @property
    def sandbox(self) -> bool:
        """Check if client use sandbox LiqPay API"""
        return is_sandbox(self._public_key)

    def update_keys(
        self, /, *, public_key: str | None, private_key: str | None
    ) -> None:
        """Update public and private keys"""
        if public_key is None:
            public_key = environ["LIQPAY_PUBLIC_KEY"]

        if private_key is None:
            private_key = environ["LIQPAY_PRIVATE_KEY"]

        sandbox = is_sandbox(public_key)
        if sandbox != is_sandbox(private_key):
            raise ValueError("Public and private keys must be both sandbox or both not")

        self._public_key = public_key
        self._private_key = secret(private_key.encode())

        warn(
            "Using %s LiqPay API" % ("sandbox" if sandbox else "live"),
            stacklevel=2,
            category=LiqpyWarning,
        )

    def __repr__(self):
        return f'{self.__class__.__name__}(public_key="{self._public_key}")'

    def _callback(
        self, /, data: bytes, signature: bytes, *, verify: bool = True
    ) -> "LiqpayCallbackDict":
        if verify:
            self.verify(data, signature)
        else:
            warn(
                "Skipping LiqPay signature verification",
                stacklevel=2,
                category=LiqpyWarning,
            )

        return decode(data, decoder=self.decoder)

    def sign(self, data: bytes, /) -> bytes:
        """
        Sign data string with private key

        See `liqpy.api.sign` for more information.
        """
        with self._private_key.dangerous_reveal() as pk:
            return sign(data, key=pk)

    def encode(
        self, /, action: str, **kwargs: Unpack["LiqpayRequestDict"]
    ) -> tuple[bytes, bytes]:
        """
        Encode parameters into data and signature strings

        >>> data, signature = client.encode("status", order_id="a1a1a1a1")

        See `liqpy.api.encode` for more information.
        """
        data = encode(
            request(action, public_key=self._public_key, version=VERSION, **kwargs),
            filter_none=True,
            validator=self.validator,
            preprocessor=self.preprocessor,
            encoder=self.encoder,
        )
        signature = self.sign(data)

        return data, signature

    def is_valid(self, /, data: bytes, signature: bytes) -> bool:
        """
        Check if the signature is valid

        Used for verification in `liqpy.Client.verify`.
        """
        return self.sign(data) == signature

    def verify(self, /, data: bytes, signature: bytes) -> None:
        """
        Verify data signature

        Raises an `AssertionError` if `data` does not match the `signature`.

        Used for verification in `liqpy.Client.callback`.
        """
        assert self.is_valid(data, signature), "Invalid signature"

    def _handle_response(self, response: Response, action: str) -> dict:
        """
        Handle the response from the LiqPay API, detecting errors and returning data.
        """
        if not response.headers.get("Content-Type", "").startswith("application/json"):
            raise exception(response=response)

        data: dict = self.decoder.decode(response.text)

        result: Optional[Literal["ok", "error"]] = data.pop("result", None)
        status = data.get("status")
        err_code = data.pop("err_code", None) or data.pop("code", None)

        if result == "ok" and status != "error":
            return data

        if action in ("status", "data") and data.get("payment_id") is not None:
            return data

        if status in ("error", "failure") or result == "error":
            raise exception(
                code=err_code,
                description=data.pop("err_description", None),
                response=response,
                details=data,
            )

        return data

    def _handle_reports_response(self, response: Response, format: Optional["Format"]) -> str:
        """
        Handle the response from the reports request, extracting data or raising errors.
        """
        output: str = response.text
        error: dict | None = None

        content_type = response.headers.get("Content-Type", "")

        if content_type.startswith("application/json"):
            if format == "json" or format is None:
                s = search(r'"data":(\[(.*)\])\}$', output)
                if s is not None:
                    output = s.group(1)
                else:
                    error = response.json()
            else:
                error = response.json()

        if error is None:
            return output
        else:
            raise exception(
                code=error.pop("err_code", None) or error.pop("code", None),
                description=error.pop("err_description", ""),
                response=response,
                details=error,
            )

    def callback(self, /, data: AnyStr, signature: AnyStr, *, verify: bool = True):
        """
        Verify and decode the callback data

        Example:
        >>> client = Client()
        >>> # get data and signature from webhook request body
        >>> order_id = "a1a1a1a1"
        >>> data, signature = client.encode(
        ...     action="pay",
        ...     amount=1,
        ...     order_id=order_id,
        ...     description="Test Encoding",
        ...     currency="USD",
        ... )
        >>> # verify and decode data
        >>> result = client.callback(data, signature)
        >>> assert result["order_id"] == order_id
        >>> assert result["action"] == "pay"
        >>> assert result["amount"] == 1
        >>> assert result["currency"] == "USD"
        >>> assert result["description"] == "Test Encoding"

        [Documentation](https://www.liqpay.ua/en/documentation/api/callback)
        """
        if isinstance(data, str):
            data = data.encode()

        if isinstance(signature, str):
            signature = signature.encode()

        result = self._callback(data, signature, verify=verify)
        version = result.get("version")

        if version != VERSION:
            logger.warning("Callback version mismatch: %s != %s", version, VERSION)

        return result

    def _handle_checkout_response(self, response: Response) -> str:
        """
        Handle the response for the checkout request, extracting the URL or raising errors.
        """
        if response.next_request is None:
            result = {}
            if response.headers.get("Content-Type", "").startswith("application/json"):
                result = response.json()

            raise exception(
                code=result.pop("err_code", None),
                description=result.pop("err_description", None),
                response=response,
                details=result if len(result) else None,
            )
        else:
            return str(response.next_request.url)


class Client(BaseClient):
    """
    Synchronous LiqPay API client using httpx.Client.
    """
    _client: _Client

    def __init__(
        self,
        /,
        public_key: str | None = None,
        private_key: str | None = None,
        *,
        client: _Client | None = None,
        **kwargs,
    ):
        super().__init__(public_key=public_key, private_key=private_key, **kwargs)
        self._client = client or _Client()

    def request(self, /, action: "Action", **kwargs: "LiqpayRequestDict") -> dict:
        response = post(
            Endpoint.REQUEST,
            *self.encode(action, **kwargs),
            client=self._client,
        )
        response.raise_for_status()  # Handle status check here
        return self._handle_response(response, action)

    def reports(
        self,
        /,
        date_from: Union[datetime, str, int, timedelta],
        date_to: Union[datetime, str, int, timedelta],
        *,
        format: Optional["Format"] = None,
    ) -> str:
        response = post(
            Endpoint.REQUEST,
            *self.encode(
                "reports", date_from=date_from, date_to=date_to, resp_format=format
            ),
            client=self._client,
        ).raise_for_status()
        return self._handle_reports_response(response, format)

    def pay(
        self,
        /,
        amount: Number,
        order_id: str | UUID,
        card: str,
        card_cvv: str,
        card_exp_month: str,
        card_exp_year: str,
        currency: "Currency",
        description: str,
        **kwargs: "LiqpayRequestDict",
    ) -> "LiqpayCallbackDict":
        """
        Request a `pay` action from LiqPay API

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/pay/doc)
        """
        return self.request(
            "pay",
            order_id=order_id,
            amount=amount,
            card=card,
            card_cvv=card_cvv,
            card_exp_month=card_exp_month,
            card_exp_year=card_exp_year,
            currency=currency,
            description=description,
            **kwargs,
        )

    def hold(
        self,
        /,
        amount: Number,
        order_id: str | UUID,
        card: str,
        card_cvv: str,
        card_exp_month: str,
        card_exp_year: str,
        currency: "Currency",
        description: str,
        **kwargs: "LiqpayRequestDict",
    ) -> "LiqpayCallbackDict":
        """
        Request a `hold` action from LiqPay API

        Use `liqpy.client.Client.complete` to complete the hold.

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/hold/doc)
        """
        return self.request(
            "hold",
            order_id=order_id,
            amount=amount,
            card=card,
            card_cvv=card_cvv,
            card_exp_month=card_exp_month,
            card_exp_year=card_exp_year,
            currency=currency,
            description=description,
            **kwargs,
        )

    def unsubscribe(self, /, opid: int | str | UUID) -> "LiqpayCallbackDict":
        """
        Cancel recurring payments for the order

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/unsubscribe/doc)
        """
        return self.request("unsubscribe", opid=opid)

    def refund(
        self,
        /,
        opid: int | str | UUID,
        *,
        amount: Number | None = None,
    ) -> "LiqpayRefundDict":
        """
        Make a refund request to LiqPay API

        Use `payment_id` (`int` type) to refund from recurring payments.
        """
        return self.request("refund", opid=opid, amount=amount)

    def complete(
        self, /, opid: int | str | UUID, *, amount: Number | None = None
    ) -> "LiqpayCallbackDict":
        """
        Request a `hold_completion` action from LiqPay API

        Use `liqpy.client.Client.hold` to request a hold action.

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/hold_completion/doc)
        """
        return self.request("hold_completion", opid=opid, amount=amount)

    def checkout(
        self,
        /,
        action: Literal["auth", "pay", "hold", "subscribe", "paydonate"],
        *,
        order_id: str | UUID,
        amount: Number,
        currency: "Currency",
        description: str,
        expired_date: str | datetime | timedelta | None = None,
        paytypes: Optional[list["PayOption"]] = None,
        **kwargs: Unpack["LiqpayRequestDict"],
    ) -> str:
        """
        Make a Client-Server checkout request to LiqPay API

        Returns a url to redirect the user to.

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/checkout/doc)
        """
        assert (
            action in CHECKOUT_ACTIONS
        ), "Invalid action. Must be one of: %s" % ",".join(CHECKOUT_ACTIONS)

        response = post(
            Endpoint.CHECKOUT,
            *self.encode(
                action,
                order_id=order_id,
                amount=amount,
                currency=currency,
                description=description,
                expired_date=expired_date,
                paytypes=paytypes,
                **kwargs,
            ),
            client=self._client,
        )
        return self._handle_checkout_response(response)

    def payments(
        self,
        /,
        date_from: Union[datetime, str, int, timedelta],
        date_to: Union[datetime, str, int, timedelta],
    ) -> list["LiqpayCallbackDict"]:
        """
        Get an archive of recieved payments

        For a significant amount of data use `liqpy.client.Client.reports` with `csv` format instead.
        """
        result = self.reports(date_from, date_to, format="json")
        return Decoder().decode(result)

    def subscribe(
        self,
        /,
        order_id: str | UUID,
        amount: Number,
        card: str,
        card_cvv: str,
        card_exp_month: str,
        card_exp_year: str,
        currency: "Currency",
        description: str,
        subscribe_periodicity: "SubscribePeriodicity",
        subscribe_date_start: datetime | str | timedelta | None | Number,
        **kwargs: Unpack["LiqpayRequestDict"],
    ) -> "LiqpayCallbackDict":
        """
        Create an order with recurring payment

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/subscribe/doc)
        """
        return self.request(
            "subscribe",
            order_id=order_id,
            amount=amount,
            card=card,
            card_cvv=card_cvv,
            card_exp_month=card_exp_month,
            card_exp_year=card_exp_year,
            currency=currency,
            description=description,
            subscribe_date_start=subscribe_date_start,
            subscribe_periodicity=subscribe_periodicity,
            **kwargs,
        )

    def subscription(
        self,
        /,
        order_id: str | UUID,
        *,
        amount: Number,
        currency: "Currency",
        description: str,
        **kwargs: Unpack["LiqpayRequestDict"],
    ) -> "LiqpayCallbackDict":
        """
        Edit an existing recurring payment

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/subscribe_update/doc)
        """
        return self.request(
            "subscribe_update",
            order_id=order_id,
            amount=amount,
            currency=currency,
            description=description,
            **kwargs,
        )

    def data(self, /, opid: str | int | UUID, *, info: str) -> "LiqpayCallbackDict":
        """
        Adding an info to already created payment

        [Documentation](https://www.liqpay.ua/en/documentation/api/information/data/doc)
        """
        return self.request("data", opid=opid, info=info)

    def ticket(
        self,
        /,
        order_id: str | UUID,
        email: str,
        *,
        payment_id: Optional[int] = None,
        language: Optional["Language"] = None,
    ) -> None:
        """
        Send a receipt to the customer

        [Documentation](https://www.liqpay.ua/en/documentation/api/information/ticket/doc)
        """
        self.request(
            "ticket",
            order_id=order_id,
            email=email,
            payment_id=payment_id,
            language=language,
        )

    def status(self, opid: int | str | UUID, /) -> "LiqpayCallbackDict":
        """
        Get the status of a payment

        [Documentation](https://www.liqpay.ua/en/documentation/api/information/status/doc)
        """
        return self.request("status", opid=opid)


class AsyncClient(BaseClient):
    """
    Asynchronous LiqPay API client using httpx.AsyncClient.
    """
    _client: _AsyncClient

    def __init__(
        self,
        /,
        public_key: str | None = None,
        private_key: str | None = None,
        *,
        client: _AsyncClient | None = None,
        **kwargs,
    ):
        super().__init__(public_key=public_key, private_key=private_key, **kwargs)
        self._client = client or _AsyncClient()

    async def request(self, /, action: "Action", **kwargs: "LiqpayRequestDict") -> dict:
        response = await post_async(
            Endpoint.REQUEST,
            *self.encode(action, **kwargs),
            client=self._client,
        )
        response.raise_for_status()  # Handle status check here
        return self._handle_response(response, action)

    async def reports(
        self,
        /,
        date_from: Union[datetime, str, int, timedelta],
        date_to: Union[datetime, str, int, timedelta],
        *,
        format: Optional["Format"] = None,
    ) -> str:
        response = await post_async(
            Endpoint.REQUEST,
            *self.encode(
                "reports", date_from=date_from, date_to=date_to, resp_format=format
            ),
            client=self._client,
        )
        response.raise_for_status()
        return self._handle_reports_response(response, format)

    async def pay(
        self,
        /,
        amount: Number,
        order_id: str | UUID,
        card: str,
        card_cvv: str,
        card_exp_month: str,
        card_exp_year: str,
        currency: "Currency",
        description: str,
        **kwargs: "LiqpayRequestDict",
    ) -> "LiqpayCallbackDict":
        """
        Request a `pay` action from LiqPay API

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/pay/doc)
        """
        return await self.request(
            "pay",
            order_id=order_id,
            amount=amount,
            card=card,
            card_cvv=card_cvv,
            card_exp_month=card_exp_month,
            card_exp_year=card_exp_year,
            currency=currency,
            description=description,
            **kwargs,
        )

    async def hold(
        self,
        /,
        amount: Number,
        order_id: str | UUID,
        card: str,
        card_cvv: str,
        card_exp_month: str,
        card_exp_year: str,
        currency: "Currency",
        description: str,
        **kwargs: "LiqpayRequestDict",
    ) -> "LiqpayCallbackDict":
        """
        Request a `hold` action from LiqPay API

        Use `liqpy.client.Client.complete` to complete the hold.

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/hold/doc)
        """
        return await self.request(
            "hold",
            order_id=order_id,
            amount=amount,
            card=card,
            card_cvv=card_cvv,
            card_exp_month=card_exp_month,
            card_exp_year=card_exp_year,
            currency=currency,
            description=description,
            **kwargs,
        )

    async def unsubscribe(self, /, opid: int | str | UUID) -> "LiqpayCallbackDict":
        """
        Cancel recurring payments for the order

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/unsubscribe/doc)
        """
        return await self.request("unsubscribe", opid=opid)

    async def refund(
        self,
        /,
        opid: int | str | UUID,
        *,
        amount: Number | None = None,
    ) -> "LiqpayRefundDict":
        """
        Make a refund request to LiqPay API

        Use `payment_id` (`int` type) to refund from recurring payments.
        """
        return await self.request("refund", opid=opid, amount=amount)

    async def complete(
        self, /, opid: int | str | UUID, *, amount: Number | None = None
    ) -> "LiqpayCallbackDict":
        """
        Request a `hold_completion` action from LiqPay API

        Use `liqpy.client.Client.hold` to request a hold action.

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/hold_completion/doc)
        """
        return await self.request("hold_completion", opid=opid, amount=amount)

    async def checkout(
        self,
        /,
        action: Literal["auth", "pay", "hold", "subscribe", "paydonate"],
        *,
        order_id: str | UUID,
        amount: Number,
        currency: "Currency",
        description: str,
        expired_date: str | datetime | timedelta | None = None,
        paytypes: Optional[list["PayOption"]] = None,
        **kwargs: Unpack["LiqpayRequestDict"],
    ) -> str:
        """
        Make a Client-Server checkout request to LiqPay API

        Returns a url to redirect the user to.

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/checkout/doc)
        """
        assert (
            action in CHECKOUT_ACTIONS
        ), "Invalid action. Must be one of: %s" % ",".join(CHECKOUT_ACTIONS)

        response = await post_async(
            Endpoint.CHECKOUT,
            *self.encode(
                action,
                order_id=order_id,
                amount=amount,
                currency=currency,
                description=description,
                expired_date=expired_date,
                paytypes=paytypes,
                **kwargs,
            ),
            client=self._client,
        )
        return self._handle_checkout_response(response)

    async def payments(
        self,
        /,
        date_from: Union[datetime, str, int, timedelta],
        date_to: Union[datetime, str, int, timedelta],
    ) -> list["LiqpayCallbackDict"]:
        """
        Get an archive of recieved payments

        For a significant amount of data use `liqpy.client.Client.reports` with `csv` format instead.
        """
        result = await self.reports(date_from, date_to, format="json")
        return Decoder().decode(result)

    async def subscribe(
        self,
        /,
        order_id: str | UUID,
        amount: Number,
        card: str,
        card_cvv: str,
        card_exp_month: str,
        card_exp_year: str,
        currency: "Currency",
        description: str,
        subscribe_periodicity: "SubscribePeriodicity",
        subscribe_date_start: datetime | str | timedelta | None | Number,
        **kwargs: Unpack["LiqpayRequestDict"],
    ) -> "LiqpayCallbackDict":
        """
        Create an order with recurring payment

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/subscribe/doc)
        """
        return await self.request(
            "subscribe",
            order_id=order_id,
            amount=amount,
            card=card,
            card_cvv=card_cvv,
            card_exp_month=card_exp_month,
            card_exp_year=card_exp_year,
            currency=currency,
            description=description,
            subscribe_date_start=subscribe_date_start,
            subscribe_periodicity=subscribe_periodicity,
            **kwargs,
        )

    async def subscription(
        self,
        /,
        order_id: str | UUID,
        *,
        amount: Number,
        currency: "Currency",
        description: str,
        **kwargs: Unpack["LiqpayRequestDict"],
    ) -> "LiqpayCallbackDict":
        """
        Edit an existing recurring payment

        [Documentation](https://www.liqpay.ua/en/documentation/api/aquiring/subscribe_update/doc)
        """
        return await self.request(
            "subscribe_update",
            order_id=order_id,
            amount=amount,
            currency=currency,
            description=description,
            **kwargs,
        )

    async def data(self, /, opid: str | int | UUID, *, info: str) -> "LiqpayCallbackDict":
        """
        Adding an info to already created payment

        [Documentation](https://www.liqpay.ua/en/documentation/api/information/data/doc)
        """
        return await self.request("data", opid=opid, info=info)

    async def ticket(
        self,
        /,
        order_id: str | UUID,
        email: str,
        *,
        payment_id: Optional[int] = None,
        language: Optional["Language"] = None,
    ) -> None:
        """
        Send a receipt to the customer

        [Documentation](https://www.liqpay.ua/en/documentation/api/information/ticket/doc)
        """
        await self.request(
            "ticket",
            order_id=order_id,
            email=email,
            payment_id=payment_id,
            language=language,
        )

    async def status(self, opid: int | str | UUID, /) -> "LiqpayCallbackDict":
        """
        Get the status of a payment

        [Documentation](https://www.liqpay.ua/en/documentation/api/information/status/doc)
        """
        return await self.request("status", opid=opid)



from typing import overload, TYPE_CHECKING
from functools import singledispatch
from numbers import Number
from datetime import datetime, timedelta, date, UTC


def from_milliseconds(value: int, tz=UTC) -> datetime:
    return datetime.fromtimestamp(value / 1000, tz=tz)


@singledispatch
def to_date(value, **kwargs) -> date:
    raise NotImplementedError(f"Unsupported type: {type(value)}")


@to_date.register
def _(value: datetime, **kwargs):
    return value.date()


@to_date.register
def _(value: date, **kwargs):
    return value


@to_date.register
def _(value: str, **kwargs):
    return date.fromisoformat(value)


@to_date.register
def _(value: timedelta, **kwargs):
    return date.today() + value


@to_date.register
def _(value: Number, **kwargs):
    return to_date(date.fromtimestamp(float(value)))


@singledispatch
def to_datetime(value, **kwargs) -> datetime:
    raise NotImplementedError(f"Unsupported type: {type(value)}")


@to_datetime.register
def _(value: datetime, **kwargs):
    return value


@to_datetime.register
def _(value: str, **kwargs):
    return datetime.fromisoformat(value)


@to_datetime.register
def _(value: Number, **kwargs):
    return datetime.fromtimestamp(float(value))


@to_datetime.register
def _(value: timedelta, **kwargs):
    return datetime.now() + value


@singledispatch
def to_milliseconds(value, **kwargs) -> int:
    raise NotImplementedError(f"Unsupported type: {type(value)}")


@to_milliseconds.register
def _(value: int, **kwargs):
    return value


@to_milliseconds.register
def _(value: datetime, **kwargs):
    return int(value.timestamp() * 1000)


@to_milliseconds.register
def _(value: str, **kwargs):
    return to_milliseconds(to_datetime(value, **kwargs))


@to_milliseconds.register
def _(value: timedelta, **kwargs):
    return to_milliseconds(to_datetime(value, **kwargs))


if TYPE_CHECKING:
    @overload
    def to_date(value: Number, **kwargs) -> date:
        ...

    @overload
    def to_date(value: date, **kwargs) -> date:
        ...

    @overload
    def to_date(value: str, **kwargs) -> date:
        ...

    @overload
    def to_date(value: timedelta, **kwargs) -> date:
        ...

    @overload
    def to_date(value: Number, **kwargs) -> date:
        ...

    @overload
    def to_datetime(value: datetime, **kwargs) -> datetime:
        ...

    @overload
    def to_datetime(value: str, **kwargs) -> datetime:
        ...

    @overload
    def to_datetime(value: timedelta, **kwargs) -> datetime:
        ...

    @overload
    def to_milliseconds(value: datetime, **kwargs) -> int:
        ...

    @overload
    def to_milliseconds(value: str, **kwargs) -> int:
        ...

    @overload
    def to_milliseconds(value: int, **kwargs) -> int:
        ...

    @overload
    def to_milliseconds(value: timedelta, **kwargs) -> int:
        ...



from datetime import datetime, timedelta
from json import loads

from pytest import fixture

from liqpy.api import Preprocessor
from liqpy.models.request import DetailAddenda

from tests import EXAMPLES_DIR


@fixture
def preprocessor():
    return Preprocessor()


def test_preprocess_report_timerange(preprocessor: Preprocessor):
    date_to = datetime.now()
    date_from = date_to - timedelta(days=30)

    r = {"date_to": date_to, "date_from": date_from}
    t = {k: int(v.timestamp() * 1000) for k, v in r.items()}

    preprocessor(r)
    assert r == t


def test_preprocess_dates(preprocessor: Preprocessor):
    t = {
        "expired_date": datetime.now(),
        "subscribe_date_start": datetime.now(),
        "letter_of_credit_date": datetime.now(),
    }

    r = {k: v.isoformat() for k, v in t.items()}
    preprocessor(r)
    assert r == t

    r = {k: v.timestamp() for k, v in t.items()}
    preprocessor(r)
    assert r == t


def test_preprocess_truthy(preprocessor: Preprocessor):
    r = {"subscribe": True, "letter_of_credit": True, "recurringbytoken": True}
    t = {k: 1 if v else None for k, v in r.items()}
    preprocessor(r)
    assert r == t

    r = {"subscribe": 1, "letter_of_credit": 1, "recurringbytoken": 1}
    t = r.copy()
    preprocessor(r)
    assert r == t

    r = {"subscribe": False, "letter_of_credit": False, "recurringbytoken": False}
    t = {k: 1 if v else None for k, v in r.items()}
    preprocessor(r)
    assert r == t

    r = {"subscribe": None, "letter_of_credit": None, "recurringbytoken": None}
    t = r.copy()
    preprocessor(r)
    assert r == t


def test_preprocess_verifycode(preprocessor: Preprocessor):
    r = {"verifycode": True}
    t = {"verifycode": "Y"}
    preprocessor(r)
    assert r == t

    r = {"verifycode": False}
    t = {"verifycode": None}
    preprocessor(r)
    assert r == t


def test_preprocess_detail_addenda(preprocessor: Preprocessor):
    with open(EXAMPLES_DIR / "dae.json") as f:
        data = loads(f.read())

    r = {"dae": data}
    t = {"dae": DetailAddenda.from_json(data)}
    preprocessor(r)

    assert r == t


from typing import TYPE_CHECKING, Any, AnyStr, Optional, Unpack

from enum import Enum
from uuid import UUID

from urllib.parse import urljoin, urlencode
from base64 import b64encode, b64decode
from hashlib import sha1

from datetime import datetime

from httpx import (
    Client,
    AsyncClient,
    Response,
    USE_CLIENT_DEFAULT,
    __version__ as httpx_version,
)

from liqpy import __version__ as liqpy_version
from liqpy.constants import URL, VERSION

from .encoder import Encoder, JSONEncoder, SEPARATORS
from .decoder import Decoder, JSONDecoder
from .preprocess import Preprocessor, BasePreprocessor
from .validation import Validator, BaseValidator
from .exceptions import exception

if TYPE_CHECKING:
    from httpx._types import TimeoutTypes, RequestExtensions

    from liqpy.types import LiqpayRequestDict
    from liqpy.types.action import Action


__all__ = [
    "Encoder",
    "Decoder",
    "Preprocessor",
    "Validator",
    "exception",
    "Endpoint",
    "is_sandbox",
    "post",
    "post_async",
    "sign",
    "encode",
    "decode",
    "request",
]


class Endpoint(Enum):
    """LiqPay API endpoints"""

    REQUEST: str = "/api/request"
    CHECKOUT: str = f"/api/{VERSION}/checkout"

    def url(self) -> str:
        """Return full URL for the endpoint"""
        return urljoin(URL, self.value)


def is_sandbox(key: str, /) -> bool:
    """Check if the key is a sandbox key"""
    return key.startswith("sandbox_")


_HEADERS = {
    "Content-Type": "application/x-www-form-urlencoded",
    "User-Agent": f"httpx/{httpx_version} {__package__}/{liqpy_version}",
}


def post(
    endpoint: Endpoint,
    /,
    data: AnyStr,
    signature: AnyStr,
    *,
    client: Client,
    timeout: Optional["TimeoutTypes"] = None,
    extensions: Optional["RequestExtensions"] = None,
) -> "Response":
    """
    Send POST request to LiqPay API

    See [Rules for the formation of a request for payment](https://www.liqpay.ua/en/documentation/data_signature).

    Arguments
    ---------
    - `endpoint` -- API endpoint to send request to (see `liqpy.Endpoint`)
    - `data` -- base64 encoded JSON data to send
    - `signature` -- LiqPay signature for the data
    - `session` -- `httpx.Client` instance to use

    Returns
    -------
    - `httpx.Response` instance

    Example
    -------
    >>> from httpx import Client
    >>> from liqpy.api import encode, sign, request, Endpoint
    >>> data = encode({"action": "status", "version": 3})
    >>> signature = sign(data, key=b"a4825234f4bae72a0be04eafe9e8e2bada209255")
    >>> with Client() as client:  # doctest: +SKIP
    ...     response = request(Endpoint.REQUEST, data, signature, client=client) # doctest: +SKIP
    ...     result = response.json() # doctest: +SKIP
    """
    return client.request(
        "POST",
        endpoint.url(),
        content=urlencode({"data": data, "signature": signature}).encode(),
        headers=_HEADERS,
        auth=None,
        timeout=USE_CLIENT_DEFAULT if timeout is None else timeout,
        follow_redirects=False,
        extensions=extensions,
    )


async def post_async(
    endpoint: Endpoint,
    /,
    data: AnyStr,
    signature: AnyStr,
    *,
    client: AsyncClient,
    timeout: Optional["TimeoutTypes"] = None,
    extensions: Optional["RequestExtensions"] = None,
) -> "Response":
    """
    Send POST request to LiqPay API asynchronously.

    See [Rules for the formation of a request for payment](https://www.liqpay.ua/en/documentation/data_signature).

    Arguments
    ---------
    - `endpoint` -- API endpoint to send request to (see `liqpy.Endpoint`)
    - `data` -- base64 encoded JSON data to send
    - `signature` -- LiqPay signature for the data
    - `client` -- `httpx.AsyncClient` instance to use

    Returns
    -------
    - `httpx.Response` instance

    Example
    -------
    >>> from httpx import AsyncClient
    >>> from liqpy.api import encode, sign, post_async, Endpoint
    >>> data = encode({"action": "status", "version": 3})
    >>> signature = sign(data, key=b"a4825234f4bae72a0be04eafe9e8e2bada209255")
    >>> async with AsyncClient() as client:  # doctest: +SKIP
    ...     response = await post_async(Endpoint.REQUEST, data, signature, client=client) # doctest: +SKIP
    ...     result = response.json() # doctest: +SKIP
    """
    return await client.request(
        "POST",
        endpoint.url(),
        content=urlencode({"data": data, "signature": signature}).encode(),
        headers=_HEADERS,
        auth=None,
        timeout=USE_CLIENT_DEFAULT if timeout is None else timeout,
        follow_redirects=False,
        extensions=extensions,
    )


def sign(data: bytes, /, key: bytes) -> bytes:
    """
    Sign data string with private key

    Algorithm:

    1. Concatenate the private key with the data string as `key + data + key`
    2. Calculate SHA1 hash of the concatenated string
    3. Encode the hash in base64

    >>> data = encode({"action": "status", "version": 3})
    >>> sign(data, key=b"a4825234f4bae72a0be04eafe9e8e2bada209255")
    b'qI0/snsDFB7MiYUxrqhBqX2420E='
    """
    return b64encode(sha1(key + data + key).digest())


def encode(
    params: "LiqpayRequestDict",
    /,
    *,
    filter_none: bool = True,
    encoder: Optional[JSONEncoder] = None,
    preprocessor: Optional[BasePreprocessor] = None,
    validator: Optional[BaseValidator] = None,
) -> bytes:
    """
    Encode parameters into base64 encoded JSON

    >>> encode({"action": "status", "version": 3})
    b'eyJhY3Rpb24iOiAic3RhdHVzIiwgInZlcnNpb24iOiAzfQ=='
    """
    if filter_none:
        params = {key: value for key, value in params.items() if value is not None}

    if encoder is None:
        encoder = JSONEncoder(separators=SEPARATORS)

    if preprocessor is not None:
        preprocessor(params, encoder=encoder)

    if validator is not None:
        validator(params)

    return b64encode(encoder.encode(params).encode())


def decode(data: bytes, /, decoder: Optional[JSONDecoder] = None) -> dict[str, Any]:
    """Decode base64 encoded JSON"""
    if decoder is None:
        decoder = JSONDecoder()

    return decoder.decode(b64decode(data).decode())


def request(
    action: "Action",
    /,
    public_key: str,
    *,
    version: int = VERSION,
    **params: "Unpack[LiqpayRequestDict]",
) -> "LiqpayRequestDict":
    """
    Create data dictionary for LiqPay API request

    >>> request("status", key="...", order_id="a1a1a1a1")
    {'action': 'status', 'public_key': '...', 'version': 3, 'order_id': 'a1a1a1a1'}
    """
    params.update(action=action, public_key=public_key, version=version)

    liqpay_id = params.pop("opid", None)
    if isinstance(liqpay_id, (str, UUID)):
        params.setdefault("order_id", liqpay_id)
    elif isinstance(liqpay_id, int):
        params.setdefault("payment_id", str(liqpay_id))

    match action:
        case "subscribe":
            assert (
                "subscribe_periodicity" in params
            ), "subscribe_periodicity is required"
            params["subscribe"] = True

            if params.get("subscribe_date_start") is None:
                params["subscribe_date_start"] = datetime.now()

        case "letter_of_credit":
            params["letter_of_credit"] = True

    return params



from typing import Literal


Language = Literal["uk", "en"]
Currency = Literal["UAH", "USD", "EUR"]
Format = Literal["json", "xml", "csv"]

PayType = Literal[
    "apay",
    "gpay",
    "apay_tavv",
    "gpay_tavv",
    "tavv",
]
PayOption = Literal[
    "card", "liqpay", "privat24", "masterpass", "moment_part", "cash", "invoice", "qr"
]
SubscribePeriodicity = Literal["month", "year"]



from datetime import datetime, date, UTC, timezone, timedelta
from uuid import UUID
from decimal import Decimal
from json import loads
from base64 import b64decode

from pytest import fixture

from liqpy.api import Encoder
from liqpy.models.request import DetailAddenda

from tests import EXAMPLES_DIR


@fixture
def encoder():
    return Encoder()


def test_encode_bytes(encoder: Encoder):
    assert encoder.encode(b"") == '""'
    assert encoder.encode(b"test") == '"test"'


def test_encode_date(encoder: Encoder):
    assert encoder.encode(date(2021, 1, 2)) == '"2021-01-02 00:00:00"'


def test_encode_datetime(encoder: Encoder):
    assert (
        encoder.encode(datetime(2021, 1, 2, 5, 4, 5, tzinfo=UTC))
        == '"2021-01-02 05:04:05"'
    )
    assert (
        encoder.encode(
            datetime(2021, 1, 2, 3, 4, 5, tzinfo=timezone(timedelta(hours=3)))
        )
        == '"2021-01-02 00:04:05"'
    )
    assert (
        encoder.encode(
            datetime(2021, 1, 2, 3, 4, 5, tzinfo=timezone(timedelta(hours=-3)))
        )
        == '"2021-01-02 06:04:05"'
    )


def test_encode_uuid(encoder: Encoder):
    value = "123e4567-e89b-12d3-a456-426614174000"
    assert encoder.encode(UUID(value)) == f'"{value}"'


def test_encode_decimal(encoder: Encoder):
    assert encoder.encode(Decimal(1)) == "1.0"
    assert encoder.encode(Decimal("1.0")) == "1.0"
    assert encoder.encode(Decimal(1.0)) == "1.0"
    assert encoder.encode(Decimal("0.0001")) == "0.0001"


def test_encode_dae(encoder: Encoder):
    with open(EXAMPLES_DIR / "dae.json") as f:
        data = loads(f.read())

    dae = DetailAddenda.from_json(data)

    assert loads(b64decode(encoder.encode(dae).encode()).decode()) == data



from typing import TYPE_CHECKING
from datetime import datetime
from re import fullmatch
from numbers import Number
from uuid import UUID
from urllib.parse import urlparse

from liqpy.models.request import DetailAddenda, SplitRule, FiscalItem, FiscalInfo

if TYPE_CHECKING:
    from liqpy.types.request import (
        DetailAddendaDict,
        SplitRuleDict,
        FiscalItemDict,
        FiscalInfoDict,
        LiqpayRequestDict,
    )


def noop(value, /, **kwargs):
    pass


def number(value, /):
    assert isinstance(value, Number), f"value must be a number"


def gt(value, /, *, threshold: Number):
    number(value)
    assert value > threshold, f"value must be greater than {threshold}"


def string(value, /, *, max_len: int | None = None):
    assert isinstance(value, str), f"value must be a string"
    if max_len is not None:
        assert len(value) <= max_len, f"string must be less than {max_len} characters"


def url(value, /, *, max_len: int | None = None):
    string(value, max_len=max_len)
    result = urlparse(value or "")
    assert result.scheme in (
        "http",
        "https",
    ), "Invalid URL scheme. Must be http or https."
    assert result.netloc != "", f"Must be a valid URL."


def to_dict(o: dict[str], cls: type) -> dict:
    if isinstance(o, cls):
        return o.__dict__
    elif isinstance(o, dict):
        return o
    else:
        raise AssertionError(f"Invalid object type. Must be {cls} or dict.")


def check_required(params: dict[str], keys: set[str]):
    missing = keys - params.keys()
    assert not missing, f"Missing required parameters: {missing}"


class BaseValidator:
    """Base class for LiqPay API request validator"""

    def __call__(self, o: "LiqpayRequestDict", /, **kwargs):
        for key, value in o.items():
            try:
                getattr(self, key, noop)(value, **kwargs.get(key, {}))
            except AssertionError as e:
                raise AssertionError(f"Invalid {key} parameter.") from e
            except Exception as e:
                raise Exception(f"Failed to verify {key} parameter.") from e


class Validator(BaseValidator):
    """LiqPay API request validator"""

    def version(self, value, /, **kwargs):
        assert isinstance(value, int), f"version must be an integer"
        assert value > 0, f"version must be greater than 0"

    def public_key(self, value, /, **kwargs):
        string(value)

    def action(self, value, /, **kwargs):
        string(value)

    def order_id(self, value, /, **kwargs):
        if not isinstance(value, UUID):
            assert isinstance(value, str), "order_id must be a string or UUID"
            assert len(value) <= 255, "order_id must be less than 255 characters"

    def description(self, value, /, **kwargs):
        # NOTE: API allows to request up to 49 720 characters,
        # but cuts to 2048 characters long
        string(value, max_len=2048)

    def amount(self, value, /, **kwargs):
        gt(value, threshold=0)

    def currency(self, value, /, **kwargs):
        assert value in ("USD", "UAH"), "currency must be USD or UAH"

    def expired_date(self, value, /, **kwargs):
        assert isinstance(value, datetime), "expired_date must be a datetime"

    def date_from(self, value, /, **kwargs):
        assert isinstance(value, int), "date_from must be a int"

    def date_to(self, value, /, **kwargs):
        assert isinstance(value, int), "date_to must be a int"

    def resp_format(self, value, /, **kwargs):
        assert value in (
            "json",
            "csv",
            "xml",
        ), "format must be json, csv or xml"

    def phone(self, value, /, **kwargs):
        assert fullmatch(
            r"\+?380\d{9}", value
        ), "phone must be in format +380XXXXXXXXX or 380XXXXXXXXX"

    def sender_phone(self, value, /, **kwargs):
        self.phone(value)

    def info(self, value, /, **kwargs):
        string(value)

    def language(self, value, /, **kwargs):
        assert value in ("uk", "en"), "language must be uk or en"

    def card_number(self, value, /, **kwargs):
        assert fullmatch(r"\d{16}", value), f"card must be 16 digits long"

    def card_cvv(self, value, /, **kwargs):
        assert fullmatch(r"\d{3}", value), f"cvv must be 3 digits long"

    def card_exp_year(self, value, /, **kwargs):
        assert fullmatch(
            r"(\d{2})?\d{2}", value
        ), f"exp_year must be 2 or 4 digits long"

    def card_exp_month(self, value, /, **kwargs):
        assert fullmatch(
            r"(0[1-9])|(1[0-2])", value
        ), f"exp_month must be 2 digits long and between 01 and 12"

    def subscribe(self, value, /, **kwargs):
        assert value == 1, "subscribe must be 1"

    def subscribe_periodicity(self, value, /, **kwargs):
        assert value in ("month", "year"), "subscribe_periodicity must be month or year"

    def subscribe_date_start(self, value, /, **kwargs):
        assert isinstance(value, datetime), "subscribe_date_start must be a datetime"

    def paytype(self, value, /, **kwargs):
        assert value in (
            "apay",
            "gpay",
            "apay_tavv",
            "gpay_tavv",
            "tavv",
        ), "paytype must be one of: apay, gpay, apay_tavv, gpay_tavv, tavv"

    def payoption(self, value, /, **kwargs):
        assert value in (
            "apay",
            "gpay",
            "card",
            "liqpay",
            "moment_part",
            "paypart",
            "cash",
            "invoice",
            "qr",
        ), "paytypes must be one of: apay, gpay, card, liqpay, moment_part, paypart, cash, invoice, qr"

    def paytypes(self, value, /, **kwargs):
        if isinstance(value, list):
            for i, item in enumerate(value):
                try:
                    self.payoption(item, **kwargs)
                except AssertionError as e:
                    raise AssertionError(f"Invalid paytypes element {i}.") from e

    def customer(self, value, /, **kwargs):
        string(value, max_len=100)

    def customer_user_id(self, value, /, **kwargs):
        string(value)

    def reccuringbytoken(self, value, /, **kwargs):
        assert value == "1", "reccuringbytoken must be 1"

    def dae(self, value, /, **kwargs):
        value: "DetailAddendaDict" = to_dict(value, DetailAddenda)

        try:
            string(value["air_line"], max_len=4)
            string(value["ticket_number"], max_len=15)
            string(value["passenger_name"], max_len=29)
            string(value["flight_number"], max_len=5)
            string(value["origin_city"], max_len=5)
            string(value["destination_city"], max_len=5)
        except AssertionError as e:
            raise AssertionError("Invalid dae object.") from e

    def split_rule(self, value, /, **kwargs):
        value: "SplitRuleDict" = to_dict(value, SplitRule)

        string(value["public_key"])
        gt(value["amount"], threshold=0)
        assert value["commission_payer"] in (
            "sender",
            "receiver",
        ), "commission_payer must be sender or receiver"
        url(value["server_url"])

    def split_rules(self, value, /, **kwargs):
        assert isinstance(value, list), "split_rules must be a list"
        for i, rule in enumerate(value):
            try:
                self.split_rule(rule, **kwargs)
            except AssertionError as e:
                raise AssertionError(
                    f"Invalid split_rule[{i}] object in split_rules."
                ) from e

    def fiscal_data(self, value, /, **kwargs):
        value: "FiscalItemDict" = to_dict(value, FiscalItem)

        assert isinstance(value["id"], int), "Invalid id. Must be an integer."

        for value in (value["amount"], value["cost"], value["price"]):
            gt(value, threshold=0)

    def rro_info(self, value, /, **kwargs):
        value: "FiscalInfoDict" = to_dict(value, FiscalInfo)

        assert isinstance(value["items"], list), "items in rro_info must be a list"
        assert isinstance(
            value["delivery_emails"], list
        ), "delivery_emails in rro_info must be a list"

        for i, item in enumerate(value["items"]):
            try:
                self.fiscal_data(item, **kwargs)
            except AssertionError as e:
                raise AssertionError(f"Invalid items[{i}] object in rro_info.") from e

        for i, email in enumerate(value.delivery_emails):
            try:
                string(email)
            except AssertionError as e:
                raise AssertionError(
                    f"Invalid delivery_emails[{i}] object in rro_info."
                ) from e

    def verifycode(self, value, /, **kwargs):
        assert value == "Y", "verifycode must be Y"

    def server_url(self, value, /, **kwargs):
        url(value, max_len=510)

    def result_url(self, value, /, **kwargs):
        # string(value, max_len=510)
        url(value, max_len=510)

    def product_url(self, value, /, **kwargs):
        url(value, max_len=2000)

    def product_description(self, value, /, **kwargs):
        string(value, max_len=500)

    def product_name(self, value, /, **kwargs):
        string(value, max_len=100)

    def product_category(self, value, /, **kwargs):
        string(value, max_len=25)

    def product_name(self, value, /, **kwargs):
        string(value, max_len=100)

    def product_category(self, value, /, **kwargs):
        string(value, max_len=25)
